/* Copyright (C)  1999  Transarc Corporation.  All rights reserved.
 */

/* Define VERSIONINFO resource */

#define  AFS_VERINFO_FILE_DESCRIPTION "AFS OSI DLL"

#include "AFS_component_version_number.h"
#include "..\..\config\NTVersioninfo.rc"
