/*
 * Copyright 2000, International Business Machines Corporation and others.
 * All Rights Reserved.
 * 
 * This software has been released under the terms of the IBM Public
 * License.  For details, see the LICENSE file in the top-level source
 * directory or online at http://www.openafs.org/dl/license10.html
 */

#define IDS_NEXT                        1
#define IDS_FINISH                      2
#define IDS_HELP_TITLE                  3
#define IDS_HELP_DESC                   4
#define IDS_CANCEL_DESC                 6
#define IDS_PARTITION_ALREADY_CREATED   7
#define IDS_NOT_A_FS_SERVER             8
#define IDS_ROOT_AFS_ALREADY_EXISTS     9
#define IDS_NO_PARTITION_EXISTS         10
#define IDS_ALREADY_REPLICATED          11
#define IDS_ROOT_AFS_DOESNT_EXIST       12
#define IDS_ALREADY_CONFIGURED          13
#define IDS_NOTHING_TO_CONFIGURE        14
#define IDS_EXIT                        15
#define IDS_START_FS_STEP               16
#define IDS_CANCEL_CONFIG_MSG           17
#define IDS_WIZARD_APP_TITLE            18
#define IDS_CANCEL_PENDING              19
#define IDS_CHECK_FS_CONFIG             20
#define IDS_CHECK_DB_CONFIG             21
#define IDS_CHECK_BAK_CONFIG            22
#define IDS_CHECK_PARTITION             23
#define IDS_CHECK_ROOT_AFS              24
#define IDS_CHECK_REP                   25
#define IDS_CANCEL_GET_CONFIG           26
#define IDS_ALREADY_A_BACKUP_SERVER     27
#define IDS_NOT_A_DB_SERVER             28
#define IDS_ALREADY_A_SYS_CONTROL_SERVER 29
#define IDS_ALREADY_A_SYS_CONTROL_CLIENT 30
#define IDS_SC_NOT_A_DB_OR_FS_SERVER    31
#define IDS_DRIVE                       32
#define IDS_NAME_OR_ERROR               33
#define IDS_BIN_NOT_A_DB_OR_FS_SERVER   34
#define IDS_CHECK_SCS                   35
#define IDS_CHECK_SCC                   36
#define IDS_SIZE                        37
#define IDS_ERROR_SEP                   38
#define IDS_MUST_CONFIG_FS              39
#define IDS_MUST_CONFIG_DB              40
#define IDS_MUST_MAKE_PARTITION         41
#define IDS_MUST_CREATE_ROOT_AFS        42
#define IDS_MUST_REPLICATE              43
#define IDS_CHECK_AFS_CLIENT            44
#define IDS_ERROR_HOST_NAME             45
#define IDS_ERROR_DRIVE_COMPRESSED      46
#define IDS_ERROR_CFG_HOST_OPEN         47
#define IDS_ERROR_AFS_CLIENT_NOT_INSTALLED 49
#define IDS_ERROR_AFS_CLIENT_CHECK      50
#define IDS_ERROR_FS_IS_NOT_NTFS        51
#define IDS_ERROR_DRIVE_HAS_DATA        52
#define IDS_PARTITION_STEP_DESC         53
#define IDS_ERROR_DRIVE_CONTAINS_NT     54
#define IDS_DB_STEP_DESC                55
#define IDS_FS_STEP_DESC                56
#define IDS_SCC_STEP_DESC               57
#define IDS_SCS_STEP_DESC               58
#define IDS_ROOT_AFS_STEP_DESC          59
#define IDS_REP_STEP_DESC               60
#define IDS_PARTITIONS_PAGE_TITLE       61
#define IDS_CONFIGURE                   62
#define IDS_DB_AND_BK_STEP_DESC         63
#define IDS_PARTITION_STEP              64
#define IDS_CONFIG_COMMON_FUNCS         65
#define IDS_START_DB_STEP               66
#define IDS_START_SCC_STEP              67
#define IDS_START_SCS_STEP              68
#define IDS_CREATE_ROOT_AFS_STEP        69
#define IDS_REP_STEP                    70
#define IDS_START_AUTH_STEP             71
#define IDS_SERVICES_PAGE_TITLE         72
#define IDS_CONFIG_SUCCEEDED            73
#define IDS_CONFIG_CANCELED             74
#define IDS_CONFIG_FAILED               75
#define IDS_ERROR_NO_LICENSE_NUM        76
#define IDS_VERIFY_CLIENT_INSTALL_STEP  77
#define IDS_SET_SERVER_LICENSE_STEP     78
#define IDS_DEFINE_CELL_NAME_STEP       79
#define IDS_DEFINE_CELL_MEMBERSHIP_STEP 80
#define IDS_START_BOS_SERVER_STEP       81
#define IDS_START_AUTH_SERVER_STEP      82
#define IDS_CREATE_PRINCIPAL_AND_KEY_STEP 83
#define IDS_START_DB_AND_BK_STEP        84
#define IDS_CREATE_ADMIN_PRINCIPAL_STEP 85
#define IDS_START_CLIENT_STEP           86
#define IDS_SET_ROOT_ACL_STEP           87
#define IDS_CREATE_ROOT_CELL_STEP       88
#define IDS_MOUNT_ROOT_CELL_STANDARD_STEP 89
#define IDS_SET_ROOT_CELL_ACL_STEP      90
#define IDS_MOUNT_ROOT_CELL_RW_STEP     91
#define IDS_ENABLE_AUTH_CHECKING_STEP   92
#define IDS_OBTAIN_ADMIN_CREDS_STEP     93
#define IDS_RESTART_SERVERS_STEP        94
#define IDS_ALREADY_A_FS_SERVER         95
#define IDS_ALREADY_A_DB_SERVER         96
#define IDS_INTRO_PAGE                  97
#define IDS_INFO_PAGE                   98
#define IDS_FS_PAGE                     99
#define IDS_DB_PAGE                     100
#define IDI_APP_ICON                    100
#define IDD_WIZARD                      101
#define IDS_CFG_TOOL_APP_TITLE          101
#define IDS_BK_PAGE                     102
#define IDS_ROOT_AFS_PAGE               103
#define IDB_GRAPHIC_256                 104
#define IDS_NAME                        104
#define IDS_REP_PAGE                    105
#define IDI_FOLDER                      106
#define IDB_GRAPHIC_16                  107
#define IDS_SC_PAGE                     108
#define IDI_FILE                        109
#define IDD_INITIAL_CONFIG              110
#define IDS_CONFIG_PAGE                 111
#define IDD_PARTITION_PAGE              112
#define IDS_PARTITION_PAGE              113
#define IDI_UPSERVER                    114
#define IDD_DB_SERVER_PAGE              115
#define IDS_GRAPHIC_FONT                116
#define IDS_CURRENT_STEP                118
#define IDD_ROOT_VOLUMES_PAGE           119
#define IDD_GET_CURRENT_CONFIG          120
#define IDS_ADD_TO_CELLSERVDB_STEP      121
#define IDD_FILE_SERVER_PAGE            122
#define IDS_RESTART_ALL_DB_SERVERS_STEP 123
#define IDD_REPLICATION_PAGE            124
#define IDS_CONFIG_SUCCEEDED_NEED_CELLSERVDB_UPDATE 125
#define IDD_CONFIG_SERVER_PAGE          126
#define IDS_SYS_CONTROL_SERVER_ONLY_MSG 127
#define IDD_INFO_PAGE                   128
#define IDS_DONT_CONFIG_SYS_CONTROL_SERVER_MSG 129
#define IDS_CLIENT_CELL_WILL_CHANGE     130
#define IDS_INFO_PAGE2                  131
#define IDS_CONFIG_SUCCEEDED_FIRST_SERVER 132
#define IDS_CONFIG_SCS                  133
#define IDS_CONFIG_INTO_CELL_MSG        134
#define IDS_VIEW_LOG                    135
#define IDS_VIEW_LOG_ERROR              136
#define IDS_ERROR_NO_LOG_FILE           137
#define IDS_UPDATING_CELLSERVDB_HOST_SUCCEEDED 138
#define IDS_UPDATING_CELLSERVDB_HOST_FAILED 139
#define IDS_CELL_NAME_LEN_ERROR         140
#define IDS_CONFIG_CHECK_FAILED         141
#define IDS_YES                         143
#define IDS_NO                          144
#define IDD_CREATE_PARTITION            146
#define IDD_INTRO_PAGE                  148
#define IDS_FREE                        149
#define IDD_BACKUP_SERVER_PAGE          150
#define IDS_TOTAL                       151
#define IDS_DRIVE2                      152
#define IDD_SYS_CONTROL_PAGE            153
#define IDS_SERVICE_STATUS              154
#define IDS_SERVICE                     156
#define IDS_FS_RUNNING                  157
#define IDS_FS_STOPPED                  158
#define IDD_INFO_PAGE2_FIRST_SERVER     159
#define IDS_DB_RUNNING                  160
#define IDD_INFO_PAGE2_NOT_FIRST_SERVER 161
#define IDS_DB_STOPPED                  162
#define IDI_DISABLED_DISK_DRIVE         163
#define IDD_SERVICES_PAGE               164
#define IDS_BK_RUNNING                  165
#define IDI_DISK_DRIVE                  166
#define IDD_PARTITIONS_PAGE             167
#define IDS_BK_STOPPED                  168
#define IDD_SERVICES_PAGE1              168
#define IDS_SC_RUNNING                  169
#define IDI_DISK_DRIVE_WITH_WARNING     169
#define IDS_SCS_STOPPED                 169
#define IDS_SC_STOPPED                  170
#define IDS_SCC_STOPPED                 170
#define IDI_AFS_DISK_DRIVE              170
#define IDS_FS_HOW_TO_RUN               171
#define IDS_FS_HOW_TO_STOP              172
#define IDS_DB_HOW_TO_RUN               173
#define IDS_DB_HOW_TO_STOP              174
#define IDS_BK_HOW_TO_RUN               175
#define IDS_BK_HOW_TO_STOP              176
#define IDS_SC_HOW_TO_RUN               177
#define IDS_SCS_HOW_TO_RUN              177
#define IDS_SC_HOW_TO_STOP              178
#define IDS_SCS_HOW_TO_STOP             178
#define IDS_DB_PARTIAL_CONFIG           179
#define IDS_DB_DETAILS                  180
#define IDS_BK_DISABLED                 181
#define IDS_BK_ENABLE                   182
#define IDS_SC_DISABLED                 183
#define IDS_SC_ENABLE                   184
#define IDS_FS_WILL_RUN                 185
#define IDS_FS_WILL_STOP                186
#define IDS_DB_WILL_RUN                 187
#define IDS_DB_WILL_STOP                188
#define IDS_BK_WILL_RUN                 189
#define IDS_BK_WILL_STOP                190
#define IDS_SC_WILL_RUN                 191
#define IDS_SCS_WILL_RUN                191
#define IDS_SC_WILL_STOP                192
#define IDS_SCS_WILL_STOP               192
#define IDS_ERROR_DRIVE_ALREADY_HAS_AFS 193
#define IDD_ADMIN_INFO                  193
#define IDS_VOLUME_HAS_NO_NAME          194
#define IDD_CONFIG_SERVER               194
#define IDS_PARTITION_CREATED           195
#define IDS_PARTITION_EXISTS            196
#define IDS_SCS_RUNNING                 197
#define IDS_WARNING_DRIVE_HAS_RECYCLE_BIN 198
#define IDD_CELLSERVDB_ERRORS           198
#define IDS_SCC_WILL_RUN                199
#define IDS_SCC_WILL_STOP               200
#define IDS_SCC_RUNNING                 201
#define IDS_SCC_HOW_TO_RUN              202
#define IDS_SCC_HOW_TO_STOP             203
#define IDS_SCS_ENABLE                  204
#define IDS_SCC_ENABLE                  205
#define IDS_UNCONFIG_DB_STEP            206
#define IDS_UNCONFIG_DB_STEP_DESC       207
#define IDS_UNCONFIG_DB_AND_BK_STEP     208
#define IDS_UNCONFIG_DB_AND_BK_STEP_DESC 209
#define IDS_UNCONFIG_BK_STEP            210
#define IDS_UNCONFIG_BK_STEP_DESC       211
#define IDS_UNCONFIG_FS_STEP            212
#define IDS_UNCONFIG_FS_STEP_DESC       213
#define IDS_UNCONFIG_SCS_STEP           214
#define IDS_UNCONFIG_SCS_STEP_DESC      215
#define IDS_UNCONFIG_SCC_STEP           216
#define IDS_UNCONFIG_SCC_STEP_DESC      217
#define IDS_INVALIDATE_CONFIG_INFO_STEP 218
#define IDS_GET_OTHER_HOSTNAME          219
#define IDS_CLOSE                       220
#define IDS_CELLSERVDB_UPDATE_ERR_MSG   221
#define IDS_DELETE_PARTITION_PROMPT     222
#define IDS_INVALIDATE_CFG_INFO         223
#define IDS_START_BK_STEP               224
#define IDS_BK_STEP_DESC                225
#define IDS_CONFIG_ERROR                226
#define IDS_CREATE_PARTITION_ERROR      227
#define IDS_REMOVE_PARTITION_ERROR      228
#define IDS_READ_PARTITIONS_ERROR       229
#define IDS_CELLSERVDB_UPDATE_PROBLEM   230
#define IDS_GET_PARTITION_LIST_ERROR    231
#define IDS_EXPORTED                    232
#define IDS_CANT_DELETE_EXPORTED_PARTITION 233
#define IDS_CHECK_CONFIG_INFO           234
#define IDS_ERROR_CFG_HOST_CLOSE        235
#define IDS_FAILED_OPEN_CELL            236
#define IDS_FAILED_OPEN_NULL_CELL       237
#define IDS_CREATE_ROOT_VOLUMES_IF_NECESSARY_PROMPT 238
#define IDS_REP_ROOT_VOLUMES_IF_NECESSARY_PROMPT 239
#define IDS_TRANSARC_AFS_SERVER         240
#define IDS_CANT_INIT_ADMIN_LIBS        241
#define IDS_CANT_OPEN_LOG_FILE          242
#define IDS_NEED_CLIENT_INFO            243
#define IDS_NEED_SERVER_INFO            244
#define IDD_SALVAGE                     244
#define IDS_ADVANCED_OPEN               245
#define IDD_SALVAGE1                    245
#define IDS_ADVANCED_CLOSED             246
#define IDS_BOS_OPEN_FAILED             247
#define IDS_SALVAGE_ERROR               248
#define IDS_INVALID_NUM_SALVAGE_PROCESSSES 249
#define IDS_WRONG_CELL                  250
#define IDS_CELLSERVDB_UPDATE_ERRORS_ARE_IN_LOG_FILE 251
#define IDS_GET_CREDENTIALS_STEP        252
#define IDS_MUST_ENTER_SCS_NAME         253
#define IDS_NOT_SAME_CELL               254
#define IDS_LAST_DB_SERVER              255
#define IDS_CELL_IS_GONE_MUST_EXIT      256
#define IDS_CFG_MAN_CANT_CONTINUE       257
#define IDS_CONFIG_INFO_INVALIDATED     258
#define IDS_GET_TOKENS_FAILED           259
#define IDD_GET_PW                      259
#define IDS_BAD_PW                      300
#define IDS_ERROR_CLIENT_VERSION        301
#define IDS_CANT_MAP_ROOT_AFS           302
#define IDS_NO_MSG_STEP                 303
#define IDS_STOP_CLIENT_STEP            304
#define IDS_ERROR_LOCAL_HOST_NAME       305
#define IDS_GET_TOKENS_ERROR            306
#define IDS_CANT_SALVAGE_WHEN_FS_NOT_CONFIGURED 307
#define IDD_SALVAGE_RESULTS             308
#define IDS_SALVAGING                   308
#define IDS_CURRENT_SALVAGE_LOG         309
#define IDS_FINAL_SALVAGE_LOG           310
#define IDS_SALVAGE_COMPLETE            311
#define IDS_CANT_GET_SALVAGE_LOG        312
#define IDS_ERROR_SAVING_SALVAGE_LOG_TO_DISK 313
#define IDS_VALIDATION_ERROR_TEMPLATE	314
#define IDS_CELL_NAME_VALIDATION_TYPE	315
#define IDS_SERVER_NAME_VALIDATION_TYPE	316
#define IDS_PASSWORD_VALIDATION_TYPE	317
#define IDS_UID_VALIDATION_TYPE			318
#define IDS_USER_NAME_VALIDATION_TYPE	319
#define IDS_PARTITION_NAME_VALIDATION_TYPE 320
#define	IDS_VOLUME_NAME_VALIDATION_TYPE	321
#define	IDS_NUM_SALVAGE_PROCS_VALIDATION_TYPE 322
#define IDS_PATH_NAME_VALIDATION_TYPE	323
#define IDS_FILE_NAME_VALIDATION_TYPE	324
#define IDS_GENERIC_VALIDATION_TYPE		325
#define IDC_OPERATION                   900
#define IDC_PROGRESS                    901
#define IDC_OPERATION2                  902
#define IDNEXT                          1001
#define IDC_CREATE_PARTITIONS           1002
#define IDBACK                          1003
#define IDC_PARTITION_LIST              1004
#define IDC_CREATE_ROOT_VOLUMES         1006
#define IDC_DONT_CONFIG_DB_SERVER       1009
#define IDC_WIZARD_LEFTPANE             1010
#define IDC_SALVAGE                     1011
#define IDC_WIZARD_RIGHTPANE            1012
#define IDC_REFORMAT                    1013
#define IDC_PARTITION_NAME              1016
#define IDC_DRIVE_LIST                  1017
#define IDC_NAME_STATIC                 1018
#define IDC_VICEP_STATIC                1019
#define IDC_MSG                         1020
#define IDC_CONFIG_PROGRESS             1022
#define IDC_LOGO                        1023
#define IDC_PARTITION_COVER             1024
#define IDC_SHOULD_CONFIG_FILE_SERVER   1025
#define IDC_PARTITION_MSG               1026
#define IDC_DB_DETAILS                  1027
#define IDC_NO_DRIVES_NOTE              1028
#define IDC_ASK_CREATE_PARTITION        1029
#define IDC_DISK_ADMIN_NOTE             1030
#define IDC_SELECT_DRIVE                1031
#define IDC_CREATE_FS_QUESTION          1032
#define IDC_CONFIG_DB_SERVER            1033
#define IDC_DB_SERVER_QUESTION          1034
#define IDC_CREATE_PARTITION            1035
#define IDC_DONT_CREATE_PARTITION       1036
#define IDC_BK_SERVICE                  1037
#define IDC_REPLICATE                   1038
#define IDC_FILE_SERVER                 1039
#define IDC_DB_SERVER                   1040
#define IDC_PARTITION                   1041
#define IDC_ROOT_AFS_VOLUME             1042
#define IDC_DONT_CONFIG_FILE_SERVER     1043
#define IDC_REPLICATION                 1044
#define IDC_DONT_CREATE_ROOT_VOLUMES    1046
#define IDC_NAME_LABEL                  1047
#define IDC_FS_SERVICE                  1048
#define IDC_SC_SERVICE                  1049
#define IDC_ROOT_AFS_MSG                1050
#define IDC_FS_ACTION_MSG               1051
#define IDC_DB_SERVICE                  1052
#define IDC_DONT_REPLICATE              1053
#define IDC_SERVER_PW                   1054
#define IDC_DB_ACTION_MSG               1055
#define IDC_CANT_REPLICATE_MSG          1056
#define IDC_ADMIN_NAME                  1057
#define IDC_VERIFY_PW                   1058
#define IDC_BK_ACTION_MSG               1059
#define IDC_ADMIN_PW                    1060
#define IDC_SC_ACTION_MSG               1061
#define IDC_CELL_NAME                   1062
#define IDC_VERIFY_ADMIN_PW             1063
#define IDC_AFS_UID                     1065
#define IDC_HOSTNAME                    1066
#define IDC_CANT_BACKUP_MSG             1067
#define IDC_CONFIG_BACKUP_SERVER        1068
#define IDC_DONT_CONFIG_BACKUP_SERVER   1069
#define IDC_BACKUP_SERVER_QUESTION      1071
#define IDC_SYS_CONTROL_MACHINE_QUESTION 1073
#define IDC_SYS_CONTROL_CLIENT          1074
#define IDC_SC_CLIENT                   1075
#define IDC_SYS_CONTROL_SERVER          1076
#define IDC_SC_SERVER                   1077
#define IDC_DONT_CONFIGURE              1078
#define IDC_AGGR_FRAME                  1079
#define IDC_SYS_CONTROL_MACHINE         1080
#define IDC_SC_MACHINE                  1081
#define IDC_SYS_CONTROL_MACHINE_LABEL   1082
#define IDC_SC_MACHINE_LABEL            1083
#define IDC_ARGS_FRAME                  1085
#define IDC_CANT_CONFIG_MSG             1091
#define IDC_JOIN_EXISTING_CELL          1092
#define IDC_ROOT_AFS_QUESTION           1095
#define IDC_REPLICATE_QUESTION          1096
#define IDC_ALL_NEEDED_MSG              1097
#define IDC_FOLLOWING_STEPS_MSG         1098
#define IDC_STEP1                       1099
#define IDC_STEP2                       1100
#define IDC_STEP3                       1101
#define IDC_STEP4                       1102
#define IDC_STEP5                       1103
#define IDC_STEP6                       1104
#define IDC_STEP7                       1105
#define IDC_STEP8                       1106
#define IDC_CURRENT_STEP                1107
#define IDC_PERCENT_COMPLETE            1108
#define IDC_REMOVE                      1109
#define IDC_PERCENT_COMPLETE_LABEL      1110
#define IDC_CURRENT_STEP_LABEL          1111
#define IDC_STATUS_MSG                  1112
#define IDC_STEP1_GRAPHIC               1113
#define IDC_STEP2_GRAPHIC               1114
#define IDC_STEP3_GRAPHIC               1115
#define IDC_STEP4_GRAPHIC               1116
#define IDC_STEP5_GRAPHIC               1117
#define IDC_STEP6_GRAPHIC               1118
#define IDC_STEP7_GRAPHIC               1119
#define IDC_STEP8_GRAPHIC               1120
#define IDC_CELL_NAME_FRAME             1121
#define IDC_SERVER_PW_FRAME             1122
#define IDC_ADMIN_INFO_FRAME            1123
#define IDC_CELL_NAME_LABEL             1124
#define IDC_CELL_NAME_PROMPT            1125
#define IDC_SERVER_PW_PROMPT            1126
#define IDC_SERVER_PW_LABEL             1127
#define IDC_ADMIN_INFO_PROMPT           1128
#define IDC_VERIFY_PW_LABEL             1129
#define IDC_ADMIN_NAME_LABEL            1130
#define IDC_CELL_NAME_PROMPT2           1131
#define IDC_ADMIN_PW_LABEL              1132
#define IDC_PRINCIPAL_LABEL             1133
#define IDC_DB_STATUS_MSG               1134
#define IDC_BK_STATUS_MSG               1136
#define IDC_SYS_CONTROL_CLIENT_DESC     1137
#define IDC_VERIFY_ADMIN_PW_LABEL       1138
#define IDC_SC_STATUS_MSG               1139
#define IDC_HOSTNAME_PROMPT             1140
#define IDC_FS_STATUS_MSG               1141
#define IDC_SCS_PROMPT                  1141
#define IDC_HOSTNAME_FRAME              1142
#define IDC_HOSTNAME_LABEL              1143
#define IDC_SCM_PROMPT                  1144
#define IDC_SCS_FRAME                   1144
#define IDC_FIRST_SERVER                1145
#define IDC_SCS_LABEL                   1145
#define IDC_AFS_UID_SPINNER             1146
#define IDC_PRINCIPAL                   1147
#define IDC_USE_THIS_UID                1148
#define IDC_USE_NEXT_UID                1149
#define IDC_TITLE                       1150
#define IDC_VALID_CONFIG_INFO           1151
#define IDC_SCS                         1152
#define IDC_SCC                         1153
#define IDC_SCM                         1154
#define IDC_BAK                         1155
#define IDC_CLIENT_INSTALLED            1156
#define IDC_FORCE_ERROR                 1157
#define IDC_ADMIN_UID                   1158
#define IDC_SALVAGE_MSG                 1159
#define IDC_REFORMAT_MSG                1160
#define IDC_REMOVE_MSG                  1161
#define IDC_CREATE                      1162
#define IDC_CLOSE                       1163
#define IDC_SC_ACTION_MSG2              1164
#define IDC_SCS_ACTION_MSG              1165
#define IDC_SCS_STATUS_MSG              1166
#define IDC_SCC_ACTION_MSG              1167
#define IDC_SCC_STATUS_MSG              1168
#define IDC_BUTTON2                     1169
#define IDC_LIST                        1171
#define IDC_CREATE_MSG                  1172
#define IDC_SERVER                      1173
#define IDC_VOLUME                      1175
#define IDC_VOLUME_NAME                 1176
#define IDC_ADVANCED                    1177
#define IDC_LOG_FILE                    1178
#define IDC_NUM_PROCESSES               1179
#define IDC_TEMP_DIR                    1180
#define IDC_TEMP_DIR_LABEL              1182
#define IDC_LOG_FILE_LABEL              1184
#define IDC_NUM_PROCESSES_CHECKBOX      1186
#define IDC_DAMAGED_VOLUMES             1187
#define IDC_SMALL_BLOCK_READS           1188
#define IDC_FORCE_SALVAGE               1189
#define IDC_FORCE_REBUILD               1190
#define IDC_LIST_DAMAGED_INODES         1191
#define IDC_LIST_OWNED_INDOES           1192
#define IDC_VOLUME_NAME_LABEL           1194
#define IDC_ADVANCED_FRAME              1195
#define IDC_PW                          1197
#define IDC_LOG                         1198
#define IDC_LOG_TITLE                   1199
#define IDC_SALVAGE_STATUS              1200

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC                     1
#define _APS_3D_CONTROLS                     1
#define _APS_NEXT_RESOURCE_VALUE        326
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1201
#define _APS_NEXT_SYMED_VALUE           102
#endif
#endif
